/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File input.c */
/* This file contains the input handling thread. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/AsciiText.h>

#include <guile/gh.h>

#include "lgd0.h"
#include "lgdlib.h"
#include "fifo.h"
#include "requests.h"

#include "ViewArea.h"

#include "input.h"
#include "pixmaps.h"
#include "output.h"
#include "legendes.h"

pthread_t input_thread;

static char *fallback_resources[] = {
  "*input: True",
  "*background: Grey51",
  "*foreground: White",
  "*groundBox.background: DarkSlateGrey",
  "*groundBox.cuteLabel.foreground: Blue",
  "*groundBox.quitButton.label: Quit",
  "*groundBox.schemeBox.submitButton.label: [enter]",
  NULL
};

Widget toplevel,ground_box,cute_label,quit_button;
Widget scheme_box,command_line,submit_button;
Widget view_area;

static XtAppContext app_context;

Display *dpy;
Screen *scn;
Window win;

Pixmap disp_pm;

/* Various C functions callable from scheme */

static SCM
cmd_modify_translations(SCM Ss)
     /* Override as specified the translation table of the ViewArea
      * widget. */
{
  char *s;
  int l;
  XtTranslations tab;
  s=gh_scm2newstr(Ss,&l);
#if 1
  tab=XtParseTranslationTable(s);
  XtOverrideTranslations(view_area,tab);
  XtFree((char *)tab); /* Am I supposed to do this? */
#else
  /* Would somebody please explain me why this fucking thing doesn't work? */
  XtVaSetValues(view_area,XtVaTypedArg,XtNtranslations,XtRString,s,l+1,NULL);
#endif
  free(s);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_move_hero(SCM Sdx,SCM Sdy)
     /* Move the hero (that is, send a move_hero request through
      * req_move_hero(). */
{
  long int dx,dy;
  dx=gh_scm2long(Sdx);
  dy=gh_scm2long(Sdy);
  req_move_hero(NULL,dx,dy);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_whereis_hero(void)
     /* Return the hero's position (that is, send a whereis_hero
      * request through reqq_whereis_hero()). */
{
  long int x,y;
  SCM Sx,Sy;
  reqq_whereis_hero(&x,&y);
  Sx=gh_long2scm(x);
  Sy=gh_long2scm(y);
  return gh_list(Sx,Sy,SCM_UNDEFINED);
}

static SCM
cmd_write_square(SCM Sx,SCM Sy,SCM Sv)
{
  long int x,y;
  u_int32_t v;
  x=gh_scm2long(Sx);
  y=gh_scm2long(Sy);
  v=gh_scm2long(Sv);
  req_write_square(NULL,x,y,v);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_write_chunk(SCM Sx,SCM Sy,SCM Sv)
{
  long int x,y;
  u_int32_t v;
  x=gh_scm2long(Sx);
  y=gh_scm2long(Sy);
  v=gh_scm2long(Sv);
  req_write_chunk(NULL,x,y,v);
  return SCM_UNSPECIFIED;
}

static SCM
cmd_read_square(SCM Sx,SCM Sy)
{
  long int x,y;
  u_int32_t v;
  x=gh_scm2long(Sx);
  y=gh_scm2long(Sy);
  reqq_read_square(x,y,&v);
  return gh_long2scm(v);
}

static SCM
cmd_read_chunk(SCM Sx,SCM Sy)
{
  long int x,y;
  u_int32_t v;
  x=gh_scm2long(Sx);
  y=gh_scm2long(Sy);
  reqq_read_chunk(x,y,&v);
  return gh_long2scm(v);
}

static SCM
cmd_sync_map(void)
{
  req_sync_map(NULL);
  return SCM_UNSPECIFIED;
}

static void
command(char *cmd)
     /* Call a scheme command. */
{
  gh_eval_str(cmd);
}

static void
quit_callback(Widget what,XtPointer client_data,XtPointer call_data)
     /* What the quit button does */
{
  exit(0);
}

static void
submit_callback(Widget what,XtPointer client_data,XtPointer call_data)
     /* What the [enter] button does */
{
  char *s,*t;
  XtVaGetValues(command_line,XtNstring,&s,NULL);
  t=lgd_malloc(1,strlen(s)+1);
  strcpy(t,s);
  XtVaSetValues(command_line,XtNstring,"",NULL);
  command(t);
  lgd_free(t);
}

static void
view_callback(Widget what,XtPointer client_data,XtPointer call_data)
     /* Called by the ViewArea widget when it dispatches some event.
      * Right now this calls the scheme function whose name is passed
      * as a parameter to the notify() action (or ``notify'' if the
      * parameter is void), with two arguments, namely the coordinates
      * of the pointer's position when the event occured (at least, if
      * the event is of KeyPress, KeyRelease, ButtonPress or
      * ButtonRelease type). */
{
  ViewAreaNotifyInfo *info = (ViewAreaNotifyInfo *)call_data;
  char t[1000]; /* XXX - THIS IS BAD */
  if (*info->num_params>=1)
    sprintf(t,"(%s %d %d)",info->params[0],
	    disp_x+info->x/ICONSIZE,disp_y+info->y/ICONSIZE);
  else
    sprintf(t,"(notify %d %d)",
	    disp_x+info->x/ICONSIZE,disp_y+info->y/ICONSIZE);
  command(t);
}

void
init_20_input(void)
{
  /* Initialize thread support (Xlib and X toolkit) */
  if (XInitThreads()==0)
    lgd_fatal("This Xlib does not support threads!");
  if (XtToolkitThreadInitialize()!=True)
    lgd_fatal("This Xt does not support threads!");
  /* Use the default language procedure (?) */
  XtSetLanguageProc(NULL,(XtLanguageProc)NULL,NULL);
  /* Create the widgets */
  toplevel = XtVaAppInitialize(&app_context,APP_CLASS,
			       NULL,0,&argc,argv,
			       fallback_resources,
			       NULL);
  ground_box = XtVaCreateManagedWidget("groundBox",boxWidgetClass,
				       toplevel,
				       NULL);
  cute_label = XtVaCreateManagedWidget("cuteLabel",labelWidgetClass,
				       ground_box,
				       XtNlabel,GAMENAME,
				       NULL);
  quit_button = XtVaCreateManagedWidget("quitButton",commandWidgetClass,
					ground_box,
					NULL);
  scheme_box = XtVaCreateManagedWidget("schemeBox",boxWidgetClass,
				       ground_box,
				       NULL);
  command_line = XtVaCreateManagedWidget("commandLine",asciiTextWidgetClass,
					 scheme_box,
					 XtNeditType,XawtextEdit,
					 NULL);
  submit_button = XtVaCreateManagedWidget("submitButton",commandWidgetClass,
					  scheme_box,
					  NULL);
  view_area = XtVaCreateManagedWidget("viewArea",viewAreaWidgetClass,
				      ground_box,
				      XtNviewWidth,WIDTH*ICONSIZE,
				      XtNviewHeight,HEIGHT*ICONSIZE,
				      NULL);
  /* Specify the callbacks */
  XtAddCallback(quit_button,XtNcallback,quit_callback,NULL);
  XtAddCallback(submit_button,XtNcallback,submit_callback,NULL);
  XtAddCallback(view_area,XtNcallback,view_callback,NULL);
  /* Get the ball rolling */
  XtRealizeWidget(toplevel);
  /* Set the important variables */
  dpy=XtDisplay(view_area);
  scn=XtScreen(view_area);
  win=XtWindow(view_area);
  disp_pm=ViewAreaGetViewPixmap(view_area);
  /* Install the various functions of the scheme interpreter */
  gh_new_procedure1_0("modify-translations",cmd_modify_translations);
  gh_new_procedure2_0("move-hero",cmd_move_hero);
  gh_new_procedure0_0("whereis-hero",cmd_whereis_hero);
  gh_new_procedure3_0("write-square",cmd_write_square);
  gh_new_procedure3_0("write-chunk",cmd_write_chunk);
  gh_new_procedure2_0("read-square",cmd_read_square);
  gh_new_procedure2_0("read-chunk",cmd_read_chunk);
  gh_new_procedure0_0("sync-map",cmd_sync_map);
  /* Run the legendesrc.scm file */
  gh_eval_file("legendesrc.scm");
}

void
demand_redraw(void)
     /* Send an X Expose event so that the ViewArea widget will redraw
      * itself.  This is used by some other thread (the output thread,
      * presumably).  _Not_ by the input thread (of course). */
{
  XEvent evt;
  evt.type=Expose;
  evt.xexpose.x=0;
  evt.xexpose.y=0;
  evt.xexpose.width=512;
  evt.xexpose.height=512;
  evt.xexpose.display=dpy;
  evt.xexpose.window=win;
  evt.xexpose.count=0;
  XSendEvent(dpy,win,True,ExposureMask,&evt);
  XFlush(dpy);
}

void *
input_main(void *garbage)
{
  XtAppMainLoop(app_context);
}
