/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File output.c */
/* This file contains the output handling thread. */

#include <stdio.h>
#include <pthread.h>

#include <X11/Xlib.h>
#include <X11/Intrinsic.h>

#include "lgd0.h"
#include "fifo.h"
#include "requests.h"

#include "ViewArea.h"
#include "input.h"
#include "pixmaps.h"
#include "output.h"

pthread_t output_thread;

#define WIDTH 32
#define HEIGHT 32

/* Display coordinates of the upper-left corner */
long int disp_x,disp_y;
/* The serial numbers of the compiled pixmaps being displayed */
unsigned long disp_tab[WIDTH][HEIGHT];

/* disp_redisplay_stat is 1 when the output thread has work to do. */
char disp_redisplay_stat;
/* The following condition serves to wake up the ouput thread when it
 * receives work to do. */
pthread_cond_t disp_redisplay_cond = PTHREAD_COND_INITIALIZER;
/* The following mutex serves to protect the above variable and
 * condition. */
pthread_mutex_t disp_redisplay_mutex = PTHREAD_MUTEX_INITIALIZER;

GC output_gc;

void
demand_redisplay(void)
     /* Wake up the output thread and ask it to redisplay.  This is
      * used by some other thread (a server thread, presumably).
      * _Not_ by the output thread (of course). */
{
  pthread_mutex_lock(&disp_redisplay_mutex);
  disp_redisplay_stat=1;
  pthread_cond_broadcast(&disp_redisplay_cond);
  pthread_mutex_unlock(&disp_redisplay_mutex);
}

void
init_40_output(void)
{
  disp_redisplay_stat=1;
  output_gc=XCreateGC(dpy,disp_pm,0,NULL);
}

void
output_loop(void)
{
  int i,j;
  long int x,y;
  /* Wait for a redisplay demand */
  pthread_mutex_lock(&disp_redisplay_mutex);
  while (!disp_redisplay_stat)
    pthread_cond_wait(&disp_redisplay_cond,&disp_redisplay_mutex);
  disp_redisplay_stat=0;
  pthread_mutex_unlock(&disp_redisplay_mutex);
  /* Where is the hero? */
  reqq_whereis_hero(&x,&y);
  {
    long int nx,ny,dx,dy;
    nx=x-16;
    ny=y-16;
    dx=nx-disp_x;
    dy=ny-disp_y;
    if ((dx>=0)&&(dy>=0)) {
      for (i=0;i+dx<WIDTH;i++) for (j=0;j+dy<HEIGHT;j++)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,dx*ICONSIZE,dy*ICONSIZE,
		(WIDTH-dx)*ICONSIZE,(HEIGHT-dy)*ICONSIZE,0,0);
    } else if ((dx>=0)&&(dy<0)) {
      for (i=0;i+dx<WIDTH;i++) for (j=HEIGHT-1;j+dy>=0;j--)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,dx*ICONSIZE,0,
		(WIDTH-dx)*ICONSIZE,(HEIGHT+dy)*ICONSIZE,0,-dy*ICONSIZE);
    } else if ((dx<0)&&(dy>=0)) {
      for (i=WIDTH-1;i+dx>=0;i--) for (j=0;j+dy<HEIGHT;j++)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,0,dy*ICONSIZE,
		(WIDTH+dx)*ICONSIZE,(HEIGHT-dy)*ICONSIZE,-dx*ICONSIZE,0);
    } else if ((dx<0)&&(dy<0)) {
      for (i=WIDTH-1;i+dx>=0;i--) for (j=HEIGHT-1;j+dy>=0;j--)
	disp_tab[i][j]=disp_tab[i+dx][j+dy];
      XCopyArea(dpy,disp_pm,disp_pm,output_gc,0,0,
		(WIDTH+dx)*ICONSIZE,(HEIGHT-dy)*ICONSIZE,
		-dx*ICONSIZE,-dy*ICONSIZE);
    }
    disp_x=nx;
    disp_y=ny;
  }
  /* Now we fill in the array */
  for (i=0;i<WIDTH;i++) for (j=0;j<HEIGHT;j++) {
    u_int32_t v;
    char h;
    u_int32_t desc[3];
    Pixmap pm;
    unsigned long int serial;
    x=disp_x+i;
    y=disp_y+j;
    reqq_readmap(x,y,&v,&h);
    desc[0]=v;
    if (h) {
      desc[1]=0x80000000;
      desc[2]=0;
    } else {
      desc[1]=0;
    }
    pm=get_pixmap(desc,&serial);
    if (serial!=disp_tab[i][j]) {
      XCopyArea(dpy,pm,disp_pm,output_gc,0,0,ICONSIZE,ICONSIZE,
		i*ICONSIZE,j*ICONSIZE);
      disp_tab[i][j]=serial;
    }
  }
  demand_redraw();
}

void *
output_main(void *garbage)
{
  while (1)
    output_loop();
}
