/* This file is part of Legendes II */
/* Copyright (c) 1998 by
 *   Philippe Clavel (philippe.clavel@supelec.fr)
 *   David Madore (david.madore@ens.fr) */

/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version
 * 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See
 * the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* File pixmaps.c */
/* This file contains the routines used to produce output pixmaps. */

#include <stdio.h>

#include <X11/Xlib.h>

#include "lgd0.h"
#include "config.h"
#include "lgdlib.h"

#include "input.h"
#include "pixmaps.h"

#define PIXMAP_TABLE_SIZE 1024

typedef struct pixmap_table_entry_s {
  Pixmap pm;
  char pm_exists;
  u_int32_t *desc;
  unsigned long int serial;
} pixmap_table_entry_t;

pixmap_table_entry_t pixmap_table[PIXMAP_TABLE_SIZE];

static unsigned long int next_serial = 0x100;

static GC pixmap_gc;

static Pixel
color_to_pixel(unsigned int red,unsigned int green,unsigned int blue)
     /* red, green, blue are in the range 0-255 */
{
  XColor color;
  color.red=red<<8;
  color.green=green<<8;
  color.blue=blue<<8;
  if (XAllocColor(dpy,DefaultColormapOfScreen(scn),&color)==0)
    lgd_fatal("Uh, oh.  I'm color starved.");
  return color.pixel;
}

static void
read_icon(u_int32_t num,int x,int y,
	  unsigned int *r,unsigned int *g,unsigned int *b,
	  unsigned int *a)
{
  if (num==0) { *r=0; *g=0; *b=0; *a=0; }
  else if (num==1) { *r=0; *g=0; *b=255; *a=0; }
  else if (num==2) { *r=0; *g=255; *b=0; *a=0; }
  else if (num==3) { *r=0; *g=255; *b=255; *a=0; }
  else if (num==4) { *r=255; *g=0; *b=0; *a=0; }
  else if (num==5) { *r=255; *g=0; *b=255; *a=0; }
  else if (num==6) { *r=255; *g=255; *b=0; *a=0; }
  else if (num==7) { *r=255; *g=255; *b=255; *a=0; }
  else if (num==0x80000000) {
    if ((x>=4)&&(x<12)&&(y>=4)&&(y<12))
      { *r=255; *g=192; *b=128; *a=0; }
    else if ((x>=2)&&(x<14)&&(y>=2)&&(y<14))
      { *r=127; *g=96; *b=64; *a=127; }
    else
      { *r=0; *g=0; *b=0; *a=255; }
  }
}

static void
fill_pixmap(int n)
     /* Draw the n-th pixmap in the pixmap table according to the
      * description already present there. */
{
  int x,y;
  int i;
  unsigned int red,green,blue;
  unsigned int r1,g1,b1,a1;
  if (!pixmap_table[n].pm_exists) {
    pixmap_table[n].pm=XCreatePixmap(dpy,win,ICONSIZE,ICONSIZE,
				     DefaultDepthOfScreen(scn));
    pixmap_table[n].pm_exists=1;
  }
  pixmap_table[n].serial=next_serial++;
  for (x=0;x<ICONSIZE;x++) for (y=0;y<ICONSIZE;y++) {
    red=0; green=0; blue=0;
    for (i=0;pixmap_table[n].desc[i];i++) {
      read_icon(pixmap_table[n].desc[i],x,y,&r1,&g1,&b1,&a1);
      red=r1+(red*a1/255);
      green=g1+(green*a1/255);
      blue=b1+(blue*a1/255);
    }
    if (red>255) red=255;
    if (green>255) green=255;
    if (blue>255) blue=255;
    XSetForeground(dpy,pixmap_gc,color_to_pixel(red,green,blue));
    XDrawPoint(dpy,pixmap_table[n].pm,pixmap_gc,x,y);
  }
}

int
desc_cmp(u_int32_t *desc1,u_int32_t *desc2)
{
  int i;
  if ((desc1==NULL)&&(desc2==NULL)) return 0;
  if (desc1==NULL) return -1;
  if (desc2==NULL) return 1;
  for (i=0;desc1[i]||desc2[i];i++) {
    if (desc1[i]<desc2[i]) return -1;
    if (desc1[i]>desc2[i]) return 1;
  }
  return 0;
}

Pixmap
get_pixmap(u_int32_t *desc,unsigned long int *serial)
     /* Return a pixmap containing the compiled form of the given
      * description. */
{
  int a,b;
  Pixmap pm;
  char pm_exists;
  a=0; b=PIXMAP_TABLE_SIZE;
  while (b>a+1) {
    int c,d;
    /* Dichotomic search */
    c=(a+b)/2;
    d=desc_cmp(desc,pixmap_table[c].desc);
    if (d==0) {
      if (serial) *serial=pixmap_table[c].serial;
      return pixmap_table[c].pm;
    }
    if (d<0) b=c;
    if (d>0) a=c;
  }
  b=lgd_random(PIXMAP_TABLE_SIZE); /* To be removed from the table */
  lgd_free(pixmap_table[b].desc);
  pm=pixmap_table[b].pm;
  pm_exists=pixmap_table[b].pm_exists;
  if (b>a) {
    int i;
    for (i=b;i>a+1;i--)
      pixmap_table[i]=pixmap_table[i-1];
    a++;
  } else {
    int i;
    for (i=b;i<a;i++) {
      pixmap_table[i]=pixmap_table[i+1];
    }
  }
  pixmap_table[a].pm=pm;
  pixmap_table[a].pm_exists=pm_exists;
  for (b=0;desc[b];b++);
  b++;
  pixmap_table[a].desc=lgd_malloc(1,b*sizeof(u_int32_t));
  for (b=0;desc[b];b++) pixmap_table[a].desc[b]=desc[b];
  pixmap_table[a].desc[b]=0;
  fill_pixmap(a);
  if (serial) *serial=pixmap_table[a].serial;
  return pixmap_table[a].pm;
}

void
init_30_pixmaps(void)
{
  pixmap_gc=XCreateGC(dpy,win,0,NULL);
}
